/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * Copyright 2005-2007 Ambient Open Source Team
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: panelbuttonclass.c,v 1.12 2009/09/27 19:18:06 kiero Exp $
 */

#include "ambient.h"

/* public */
#include <cybergraphx/cybergraphics.h>
#include <proto/cybergraphics.h>
#include <proto/graphics.h>

/* private */
#include "rexx.h"
#include "command.h"
#include "contextmenu.h"
#include "iconio.h"
#include "name.h"
#include "gfx_scale.h"
#include "gfx_bitmap.h"
#include "gfx_blit.h"
#include "gfx_pen.h"
#include "prefs.h"
#include "datatypes_picture.h"
#include "wbstart.h"
#include "threads.h"
#include "file_func.h"
#include "dragdrop.h"
#include "legacy.h"


enum
{
	BMEFFECT_NONE = 0,      /* XXX: from icondata.h, maybe something more global could be invented... */
	BMEFFECT_SELECTED,
	BMEFFECT_MOUSEOVER,
	BMEFFECT_DRAGDROP
};

struct Data {
	ULONG distance;
	APTR cmenu;
	STRPTR imagepath;
	STRPTR uri;
	STRPTR help;
	ULONG directory;
	#if 0
	struct MUI_EventHandlerNode ehnode;
	#endif
	APTR bm;
	APTR sclbm;
	ULONG oldwidth;
	ULONG oldheight;

	/* image highlighting */

	ULONG highlighted, pressed;
	ULONG droptarget;
	ULONG current_effect, desired_effect;
	APTR tempbm;
};

static void doset(APTR obj, struct Data *data, struct TagItem *tags)
{
	ULONG update = 0;

	FORTAG(tags)
	{
		case MA_Panel_Distance:
			data->distance = tag->ti_Data;
			break;

		case MA_Panel_Imagepath:
			if (data->imagepath)
			{
				name_delete(data->imagepath);
			}
			data->imagepath = name_build((STRPTR)tag->ti_Data);
			{
				if ( data->help )
					name_delete( data->help );

				data->help = name_build_noinfo( data->imagepath );
				if ( data->help )
					set( obj, MUIA_ShortHelp, data->help );
			}
			break;

		case MA_Panel_URI:
			if (data->uri)
			{
				name_delete(data->uri);
			}
			data->uri = name_build((STRPTR)tag->ti_Data);
			data->directory = isdir( data->uri );

			break;

		case MA_Panel_Highlighted:
			data->highlighted = tag->ti_Data;
			update = 1;
			break;

		case MUIA_Selected:
			data->pressed = tag->ti_Data;
			update = 1;
			break;
	}
	NEXTTAG

	if (data->pressed) 
	{
		data->highlighted = FALSE;
	}

	if (update)
	{
		data->desired_effect = data->droptarget ? BMEFFECT_DRAGDROP : ( data->pressed ? BMEFFECT_SELECTED : BMEFFECT_MOUSEOVER );
		MUI_Redraw(obj, MADF_DRAWOBJECT);
	}
}


DEFNEW
{
	struct Data *data;

	obj = DoSuperNew(cl, obj,
		//ButtonFrame, /* XXX: instead of button frame effect we could use other effects here... must think something cool */
		MUIA_InnerLeft, 0,
		MUIA_InnerRight, 0,
		MUIA_InnerTop, 0,
		MUIA_InnerBottom, 0,
		MUIA_InputMode, MUIV_InputMode_RelVerify,
		//MUIA_Font, ... XXX: could make sense
		MUIA_ContextMenu, 1,
/*		MUIA_FrameDynamic, TRUE,
		MUIA_FillArea, FALSE,
		MUIA_FrameVisible, FALSE, */
		MUIA_ShowSelState, FALSE, /* _conf(icon_hover) ? FALSE : TRUE,  <- doesn't work in current MUI, so just skip for now */
		TAG_MORE, INITTAGS
	);

	if (!obj)
	{
		return (NULL);
	}

	data = INST_DATA(cl, obj);

	data->help = NULL;
	data->droptarget = FALSE;

	doset(obj, data, INITTAGS);

	DoMethod(obj, MUIM_Notify, MUIA_Pressed, FALSE,
		obj, 1, MM_Panelbutton_Launch
	);

	return ((ULONG)obj);
}


DEFDISP
{
	GETDATA;

	gfx_bitmap_delete(data->sclbm);

	gfx_bitmap_delete(data->bm);

	if (data->uri)
	{
		name_delete(data->uri);
	}

	if (data->help)
	{
		name_delete(data->help);
	}

	if (data->imagepath)
	{
		name_delete(data->imagepath);
	}

	gfx_bitmap_delete(data->tempbm);

	return (DOSUPER);
}


DEFMMETHOD(AskMinMax)
{
	ULONG size = getv(_parent(obj), MA_Panelgroup_Size);

	DOSUPER;

	ASSERT(size);

	/* XXX: beware.. we don't use += here because MUI adds too much frames atm.. experiment later */
	msg->MinMaxInfo->MinWidth = size;
	msg->MinMaxInfo->MinHeight = size;
	msg->MinMaxInfo->MaxWidth = size;
	msg->MinMaxInfo->MaxHeight = size;

	return (0);
}


DEFGET
{
	switch (msg->opg_AttrID)
	{
		case MA_Panel_Type:
			*msg->opg_Storage = MV_Panel_Type_Button;
			return (TRUE);

		case MA_Panel_Distance:
			{
				GETDATA;
				*msg->opg_Storage = data->distance;
			}
			return (TRUE);

		case MA_Panel_Imagepath:
			{
				GETDATA;
				*msg->opg_Storage = (ULONG)data->imagepath;
			}
			return (TRUE);
	}
	return (DOSUPER);
}


DEFSET
{
	GETDATA;

	doset(obj, data, INITTAGS);

	return (DOSUPER);
}


DEFMMETHOD(Setup)
{
	set(obj, MUIA_ShowSelState, !_conf(icon_hover));

	return (DOSUPER);
}

#if 0
DEFMMETHOD(Setup)
{
	GETDATA;

	if (!DOSUPER)
	{
		return (0);
	}

	if (muiRenderInfo(obj) && _win(obj))
	{
		data->ehnode.ehn_Object = obj;
		data->ehnode.ehn_Class = cl;
		data->ehnode.ehn_Events = IDCMP_MOUSEOBJECT;
		data->ehnode.ehn_Priority = 1; /* XXX: dumdidum */
		data->ehnode.ehn_Flags = MUI_EHF_GUIMODE;
		DoMethod(_win(obj), MUIM_Window_AddEventHandler, (ULONG)&data->ehnode);
	}
	return (TRUE);
}
#endif

#if 0
DEFMMETHOD(HandleEvent)
{
	if (msg->imsg)
	{
		if (msg->imsg->Class == IDCMP_MOUSEOBJECT)
		{

		}
	}
	return (0);
}
#endif


#if 0
DEFMMETHOD(Cleanup)
{
	if (muiRenderInfo(obj) && _win(obj))
	{
		GETDATA;
		DoMethod(_win(obj), MUIM_Window_RemEventHandler, (ULONG)&data->ehnode);
	}
	return (DOSUPER);
}
#endif


DEFMMETHOD(ContextMenuBuild)
{
	GETDATA;

	if (data->cmenu)
	{
		MUI_DisposeObject(data->cmenu);
	}

	if ( (data->cmenu = contextmenu_build(CM_PANELBUTTON, NULL)) ) /* XXX: we should honor locked mode perhaps */
	{
		return ((ULONG)data->cmenu);
	}
	errormsg(ERR_NOMEM);

	return (NULL);
}

DEFMMETHOD(ContextMenuAdd)
{
	GETDATA;
	ULONG flags = 0;

	if (!getv(_parent(obj), MA_Panelgroup_Locked))
		flags |= AS_PANEL_UNLOCKED;

	data->cmenu	= contextmenu_build_simple(CM_PANELBUTTON, flags);

	if (data->cmenu)
		DoMethod(msg->menustrip, OM_ADDMEMBER, data->cmenu);

	return (DOSUPER);
}


DEFMMETHOD(ContextMenuChoice)
{
	GETDATA;
	struct command_menu *cm;
	ULONG rc = 0;

	if (data->cmenu && DoMethod(data->cmenu, MUIM_FindObject, msg->item))
	{
		if ( ((cm = (struct command_menu *)getv(msg->item, MA_Menuitem_Command))) )
		{
			if (cm->name && *cm->name)
				execute_command(_parent(obj), cm->type, cm->args, NULL);
		}
	}
	else
		rc = DOSUPER;


	data->cmenu = NULL;

	return (rc);
}


DEFSMETHOD(Icon_AddBitMap)
{
	GETDATA;

	data->oldwidth = 0;
	data->oldheight = 0;

	gfx_bitmap_delete( data->tempbm );
	data->tempbm = NULL;


	/*
	 * We only keep one bitmap.
	 */
	switch (msg->type)
	{
		case MV_Icon_BitMap_Standard:
			switch (msg->state)
			{
				case MV_Icon_BitMap_Normal:
					gfx_bitmap_delete(data->bm);
					data->bm = msg->bm;
					break;

				case MV_Icon_BitMap_Selected:
					gfx_bitmap_delete(msg->bm);
					break;
			}
			break;

		case MV_Icon_BitMap_Newicon:
			switch (msg->state)
			{
				case MV_Icon_BitMap_Normal:
					gfx_bitmap_delete(data->bm);
					data->bm = msg->bm;
					break;

				case MV_Icon_BitMap_Selected:
					gfx_bitmap_delete(msg->bm);
					break;
			}
			break;

		case MV_Icon_BitMap_Glowicon:
			switch (msg->state)
			{
				case MV_Icon_BitMap_Normal:
					gfx_bitmap_delete(data->bm);
					data->bm = msg->bm;
					break;

				case MV_Icon_BitMap_Selected:
					gfx_bitmap_delete(msg->bm);
					break;
			}
			break;

		#if USE_PNGICONS
		case MV_Icon_BitMap_PNGicon:
			data->bm = msg->bm;
			break;
		#endif

		#if USE_DTICONS
		case MV_Icon_BitMap_DTicon:
			data->bm = msg->bm;
			break;
		#endif

		#ifdef DEBUG
		default:
			PDB(("out of bound bitmap: val: %ld\n", msg->type));
			break;
		#endif
	}
	return (0);
}


DEFTMETHOD(Panelbutton_LoadImage)
{
	GETDATA;

	if (data->imagepath)
	{
		icon_read(data->imagepath, obj, /* XXX: argh! not in that context, launch a thread */
			ICONTAG_Position, FALSE,
			ICONTAG_Deficon, TRUE,
		TAG_DONE);

	}

	return (0);
}

DEFMMETHOD(DragQuery)
{
	if (!getv(_parent(obj), MA_Panelgroup_Locked))
	{
		STRPTR path;

		if (get(msg->obj, MA_Icon_PathInfo, &path)) /* XXX: hm, fuck.. that works for iconviews as well.. (broadcast) */
		{
			if (getv(msg->obj, MA_Icon_Type) != MV_Icon_Type_MyComputer &&
				getv(msg->obj, MA_Icon_Type) != MV_Icon_Type_AppIcon)
				return (MUIV_DragQuery_Accept);
		}
	}
	else
	{
		/*
		 * Allow files/dirs. Will be used as param.
		 */

		/* XXX: This is a bit lame check. Change to real file recognition later. */

		GETDATA;
		LONG numselected = 0;

		if ( data->directory )
			return (MUIV_DragQuery_Refuse);

		/* check for valid object being dragged! */

		if (get(msg->obj, MA_View_NumSelected, &numselected) == FALSE)
			return (MUIV_DragQuery_Refuse);

        /* i wonder if numsel == 0 still happens */

		if ( numselected == 0 )
		{
			/* single file dropped */

			LONG ft = getv(msg->obj, MA_Icon_FileType);

			if (ft == MV_Icon_FileType_File || ft == MV_Icon_FileType_Directory )
			{
				return (MUIV_DragQuery_Accept);
			}
		}
		else if ( numselected >= 1 )
		{
			/* multiple files dropped. check if all are of allowed type */
			struct MinList ml;
			struct dragdropnode *ddn, *nextddn;
			ULONG ret = MUIV_DragQuery_Accept;

			NEWLIST(&ml);

			DoMethod(_view(msg->obj), MM_View_GetSelectionList, &ml);

			ITERATELISTSAFE(ddn, nextddn, &ml)
			{
				if(ddn->type !=  MV_Icon_FileType_File && ddn->type != MV_Icon_FileType_Directory )
				{
					ret = MUIV_DragQuery_Refuse;
					/* NOTE: no break since we must free all dragdropnodes regardless */
				}
				free(ddn);
			}

			return (ret);
		}

	}

	return (MUIV_DragQuery_Refuse);
}

DEFSMETHOD(Panelbutton_AddIcon)
{
	GETDATA;
	STRPTR path;

	ASSERT(msg->obj);

	if (get(msg->obj, MA_Icon_PathInfo, &path))
	{
		if (data->imagepath)
		{
			name_delete(data->imagepath);
		}

		/*
		 * Try to put the executable's filename in
		 * there.
		 * XXX: handle more URIs later on
		 */
		if (data->uri)
		{
			name_delete(data->uri);
			data->uri = NULL;
		}

		data->uri = name_build((STRPTR)getv(msg->obj, MA_Icon_Path));
		data->directory = isdir( data->uri );

		if ( (data->imagepath = name_build(path)) )
		{
			if ( data->help )
				name_delete( data->help );

			data->help = name_build_noinfo( data->imagepath );
			if ( data->help )
			{
				set( obj, MUIA_ShortHelp, data->help );
			}

			DoMethod(obj, MM_Panelbutton_LoadImage); /* XXX: that needs to be async, then button needs to be set as 'loaded' and no action can be performed until the thread is back! (context menu, etc..) */

		}
		/* XXX */

		set(_parent(obj), MA_Panelgroup_HasChanged, TRUE);
	}
	return (0);
}

DEFSMETHOD(Thread_Finished)
{
#if 0
/* now freed in wbstart */
	switch (msg->action)
	{
		case TA_WBStart:
			{
				/*
				 * We have a list to free.
				 */

				struct MinList *l = ( struct MinList* )GetTagData(TT_WBStart_Argument_List, NULL, msg->taglist );

				if ( l )
				{
					struct wbsnode *wbsn, *nextwbsn;

					ITERATELISTSAFE( wbsn, next wbsn, l)
					{
						free( wbsn);
					}

					free( l );
				}
			}
			break;
	}
#endif
	return (0);
}

DEFMMETHOD(DragDrop)
{
	if (!getv(_parent(obj), MA_Panelgroup_Locked))
	{
		DoMethod(obj, MM_Panelbutton_AddIcon, msg->obj);
		MUI_Redraw(obj, MADF_DRAWOBJECT);
	}
	else
	{
		/*
		 * Launch with dropped file as param.
		 */

		GETDATA;

		if ( data->uri )
		{
			struct MinList *l = malloc( sizeof( struct MinList ) );

			if ( l )
			{
				int num = 0;

				NEWLIST( l );

				/*
				 * This list is deallocated (together with nodes) when thread finished.
				 */

				/* i wonder if numsel == 0 still happens */
				if ( getv(msg->obj, MA_View_NumSelected) == 0 )
				{
					STRPTR path = (STRPTR)getv( msg->obj, MA_Icon_Path );
					if ( path )
					{
						struct wbsnode *n = malloc( sizeof( struct wbsnode ) + strlen( path ) + 1 );
						if ( n )
						{
							strcpy( n->arg, path );
							ADDTAIL( l, n );
							num++;
						}
					}
				}
				else if ( getv(msg->obj, MA_View_NumSelected) >= 1 )
				{
					struct MinList ml;
					struct dragdropnode *ddn, *nextddn;

					NEWLIST(&ml);

					DoMethod(_view(msg->obj), MM_View_GetSelectionList, &ml);

					ITERATELISTSAFE(ddn, nextddn, &ml)
					{
						struct wbsnode *n = malloc( sizeof( struct wbsnode ) + strlen( ddn->path ) + 1 );
						if ( n )
						{
							strcpy( n->arg, ddn->path );
							ADDTAIL( l, n );
							num++;
						}
						/* XXX: careful here, if you break out you must still free the nodes */
						free(ddn);
					}
				}

				if ( num )
				{
					/*
					 * apparently TA_WBStart does free the argument list.
					 * check wbstart.c/wbstart(). - piru
					 */

					do_action( obj, TA_WBStart, 
										TT_WBStart_Path, data->uri,
										TT_WBStart_Argument_List, l,
									TAG_DONE);
				}
			}
		}

	}

	return (0);
}


/*
 * Prevents areaclass from drawing a stupid frame around.
 */
#if 1
DEFMMETHOD(DragBegin)
{

	GETDATA;

	if ( msg->obj != obj )
	{
		if ( getv(_parent(obj), MA_Panelgroup_Locked) )
		{
			data->droptarget = TRUE;
			data->desired_effect = BMEFFECT_DRAGDROP;
			data->current_effect = -1;
			MUI_Redraw(obj, MADF_DRAWOBJECT);
			return (0);
		}
	}

	return(DOSUPER);
}

DEFMMETHOD(DragFinish)
{
	GETDATA;

	if ( msg->obj != obj )
	{
		if ( getv(_parent(obj), MA_Panelgroup_Locked) )
		{
			data->droptarget = FALSE;
			data->desired_effect = 0;
			data->current_effect = -1;
			MUI_Redraw(obj, MADF_DRAWOBJECT);
			return (0);
		}
	}

	return (DOSUPER);
}

#endif


static VOID process_image(APTR obj, struct IClass *cl, APTR srcbm, APTR dstbm)
{
	struct RastPort rp;
	ULONG color;
	#if USE_DROP_EFFECT_PREFS
	ULONG colorfade;
	ULONG brighten;
	ULONG darken;
	ULONG effect;
	#endif

	GETDATA;

	ASSERT(dstbm);
	ASSERT(data->current_effect != data->desired_effect);

	InitRastPort(&rp);
	rp.BitMap = gfx_bitmap_bm(dstbm);

	gfx_blit(srcbm, dstbm, TAG_DONE);
	data->current_effect = data->desired_effect;

	#if USE_DROP_EFFECT_PREFS
	if ( data->droptarget )
	{
		effect = _conf(dragdrop_dropeffect);
		color = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(dragdrop_tintval));
		colorfade = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(dragdrop_tintfadeval));
		brighten = _conf(dragdrop_brightenval);
		darken = _conf(dragdrop_darkenval);

	}
	else
	{
		effect = _conf(icon_selecteffect);
		color = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(icon_tintval));
		colorfade = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(icon_tintfadeval));
		brighten = _conf(icon_brightenval);
		darken = _conf(icon_darkenval);
	}
	#else
	if ( data->droptarget )
	{
		color = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(dragdrop_tintval));
	}
	else
	{
		color = gfx_get_penspec_value(muiRenderInfo(obj), &_conf(icon_tintval));
	}
	#endif


	switch (data->desired_effect)
	{
		case BMEFFECT_SELECTED:
		case BMEFFECT_DRAGDROP:
			#if USE_DROP_EFFECT_PREFS
			switch (effect)
			#else
			switch (DE_TINT)
			#endif
			{
				case DE_TINT:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_TINT, color, NULL);
					break;

				#if USE_DROP_EFFECT_PREFS
				case DE_BLUR:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_BLUR, 0, NULL);
					break;

				case DE_BRIGHTEN:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_BRIGHTEN, brighten, NULL);
					break;

				case DE_DARKEN:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_DARKEN, darken, NULL);
					break;

				case DE_GREY:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_COLOR2GREY, 0, NULL);
					break;

				case DE_NEGATIVE:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_NEGATIVE, 0, NULL);
					break;

				case DE_NEGFADE:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_NEGFADE, 0, NULL);
					break;

				case DE_TINTFADE:
					ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_TINTFADE, colorfade, NULL);
					break;
				#endif
			}
			break;

		case BMEFFECT_MOUSEOVER:
			ProcessPixelArray(&rp, 0, 0, gfx_bitmap_width(dstbm), gfx_bitmap_height(dstbm), POP_BRIGHTEN, 0x20, NULL);
			break;
	}
}

DEFMMETHOD(Draw)
{
	ULONG rc;
	GETDATA;

	rc = DOSUPER;

	if (data->bm && msg->flags & (MADF_DRAWOBJECT|MADF_DRAWUPDATE))
	{
		APTR tbm;
		ULONG mleft, mtop, mwidth, mheight;
		ULONG txs;
		ULONG tys;
		ULONG updated = FALSE;

		mleft = _mleft(obj);
		mtop = _mtop(obj);
		mwidth = _mwidth(obj);
		mheight = _mheight(obj);

		if ((data->oldwidth != mwidth) || (data->oldheight != mheight))
		{
			updated = TRUE;

			gfx_bitmap_delete(data->sclbm);
			data->sclbm = NULL;

			ASSERT(data->bm);

			if ((gfx_bitmap_width(data->bm) > mwidth) || (gfx_bitmap_height(data->bm) > mheight))
			{
				gfx_bitmap_delete(data->tempbm);
				data->tempbm = NULL;

				/*
				 * Calculate new dimensions (preserve aspect).
				 */

				gfx_scale_calc_aspect_constraints(
					gfx_bitmap_width(data->bm),
					gfx_bitmap_height(data->bm),
					&mwidth,
					&mheight
				);

				/*
				 * We need to rescale.
				 */
				if ( (data->sclbm = gfx_bitmap_create(mwidth, mheight, 32, BITMAPTAG_Clear, TRUE, BITMAPTAG_Format, BITMAPVAL_Format_ARGB32, TAG_DONE)) )
				{
					/* XXX: hm.. fix centering.. the whole probably needs some overhault.. BMF_CLEAR above is a kludge . */
					gfx_scale(data->bm, data->sclbm, mwidth, mheight,
						SCALETAG_Scale2x, TRUE,
						SCALETAG_Bilinear, TRUE,
						SCALETAG_Average, TRUE,
						SCALETAG_AspectX, NULL,
						SCALETAG_AspectY, NULL,
					TAG_DONE);
				}
				/* XXX: hm.. what to do */
			}
			data->oldwidth = mwidth;
			data->oldheight = mheight;
		}

		if (data->sclbm)
		{
			tbm = data->sclbm;
			txs = mwidth;
			tys = mheight;
		}
		else
		{
			tbm = data->bm;
			txs = gfx_bitmap_width(data->bm);
			tys = gfx_bitmap_height(data->bm);
		}

		if (data->highlighted || data->droptarget || data->pressed)
		{
			if (data->tempbm && (gfx_bitmap_width(data->tempbm) != gfx_bitmap_width( tbm ) || gfx_bitmap_height(data->tempbm) != gfx_bitmap_height( tbm )))
			{
				/*
				 * temp bm have to be reallocated. For some obscure and hard to debug
				 * reason it's at wrong dimensions sometimes. This is more a workaround
				 * than a fix.
				 */

				gfx_bitmap_delete(data->tempbm);
				data->tempbm = NULL;
			}

			if (!data->tempbm)
			{
				data->tempbm = gfx_bitmap_create(txs, tys, 32, BITMAPTAG_Format, BITMAPVAL_Format_ARGB32, TAG_DONE);

				if (!data->tempbm)
					return rc;

				updated = TRUE;
			}

			if ( (data->desired_effect != data->current_effect) || updated )
			{
				process_image(obj, cl, tbm, data->tempbm);
			}

			tbm = data->tempbm;
		}

		gfx_blit(tbm, _rp(obj),
			BLITTAG_DstType, BLITVAL_DstType_RastPort,
			BLITTAG_DstX, mleft + (_mwidth( obj ) - txs) / 2,
			BLITTAG_DstY, mtop + (_mheight( obj ) - tys) / 2,
			BLITTAG_Alpha, 0xffffffff,
		TAG_DONE);
	}

	/* XXX: no fallback ? we should draw a frame or so */
	return (rc);
}


DEFTMETHOD(Icon_End)
{
	MUI_Redraw(obj, MADF_DRAWUPDATE);

	return (0);
}


DEFTMETHOD(Panelbutton_Launch)
{
	GETDATA;

	if (data->uri)
	{
		TEXT buf[PATH_SIZE + 256]; /* should be enough (tm) */

		snprintf(buf, sizeof(buf), "LoadURI \"%s\"", data->uri);

		execute_command(obj, AC_INTERNAL, buf, NULL);
	}
	return (0);
}


DEFSMETHOD(Panel_SaveConfig)
{
	APTR pl, pi; /* list, item */
	GETDATA;
	ASSERT(msg->pctx);

	if (!(pl = prefspool_item_get(msg->pctx, NULL, DSI_LISTPOOL_PANEL, NULL, NULL)))
	{
		pl = prefspool_item_add(msg->pctx, NULL, DSI_LISTPOOL_PANEL, NULL, NULL);
	}

	if (pl)
	{
		if (!(pi = prefspool_item_get(msg->pctx, pl, msg->index | DSF_LISTPOOL, NULL, NULL)))
		{
			pi = prefspool_item_add(msg->pctx, pl, msg->index | DSF_LISTPOOL, NULL, NULL);
		}
		if (pi)
		{
			/* XXX: we should check the retvals or so */
			setprefslong_lp(msg->pctx, pi, DSI_LISTPOOL_PANEL_DISTANCE, data->distance);

			/* XXX: perhaps we should just delete the prefs item instead of setting it to nothing */
			setprefsstr_lp(msg->pctx, pi, DSI_LISTPOOL_PANEL_IMAGEPATH, data->imagepath ? data->imagepath : (STRPTR)"");
			setprefsstr_lp(msg->pctx, pi, DSI_LISTPOOL_PANEL_URI, data->uri ? data->uri : (STRPTR)"");
		}
		/* XXX */
	}
	/* XXX */
	return (0); /* XXX: what about some useful result ? */
}


BEGINMTABLE
DECNEW
DECDISP
DECGET
DECSET
DECMMETHOD(AskMinMax)
DECMMETHOD(Setup)
//DECMMETHOD(Cleanup)
//DECMMETHOD(HandleEvent)
//DECMMETHOD(ContextMenuBuild)
DECMMETHOD(ContextMenuAdd)
DECMMETHOD(ContextMenuChoice)
DECSMETHOD(Icon_AddBitMap)
DECSMETHOD(Panelbutton_AddIcon)
DECMMETHOD(DragQuery)
DECMMETHOD(DragBegin)
DECMMETHOD(DragFinish)
DECMMETHOD(DragDrop)
DECTMETHOD(Panelbutton_LoadImage)
DECMMETHOD(Draw)
DECTMETHOD(Icon_End)
DECTMETHOD(Panelbutton_Launch)
DECSMETHOD(Panel_SaveConfig)
DECSMETHOD(Thread_Finished)

ENDMTABLE

DECSUBCLASS_NC(MUIC_Area, panelbuttonclass)
